@@ -26,18 +26,20 @@
 #include "vinagre-fav.h"
 #include "vinagre-utils.h"
 #include "vinagre-bookmarks.h"
-#include "vinagre-mdns.h"
 #include "vinagre-window-private.h"
-#include "gossip-cell-renderer-expander.h"
+#include "vinagre-bookmarks-entry.h"
+
+#ifdef VINAGRE_ENABLE_AVAHI
+#include "vinagre-mdns.h"
+#endif
  
 #define VINAGRE_FAV_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_FAV, VinagreFavPrivate))
 
 struct _VinagreFavPrivate
 {
-  VinagreWindow *window;
-  GtkWidget     *tree;
-  GtkTreeModel  *model;
-  GtkTreeViewColumn *main_column;
+  VinagreWindow     *window;
+  GtkWidget         *tree;
+  GtkTreeModel      *model;
 };
 
 G_DEFINE_TYPE(VinagreFav, vinagre_fav, GTK_TYPE_VBOX)
@@ -54,7 +56,7 @@ enum
 enum {
   IMAGE_COL = 0,
   NAME_COL,
-  CONN_COL,
+  ENTRY_COL,
   IS_FOLDER_COL,
   IS_GROUP_COL,
   IS_AVAHI_COL,
@@ -71,12 +73,6 @@ enum
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
-static void
-vinagre_fav_finalize (GObject *object)
-{
-  G_OBJECT_CLASS (vinagre_fav_parent_class)->finalize (object);
-}
-
 static void
 vinagre_fav_set_property (GObject      *object,
 			  guint         prop_id,
@@ -122,7 +118,6 @@ vinagre_fav_class_init (VinagreFavClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-  object_class->finalize = vinagre_fav_finalize;
   object_class->get_property = vinagre_fav_get_property;
   object_class->set_property = vinagre_fav_set_property;
 
@@ -141,7 +136,7 @@ vinagre_fav_class_init (VinagreFavClass *klass)
 			      G_SIGNAL_RUN_FIRST,
 			      G_STRUCT_OFFSET (VinagreFavClass, fav_activated),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__POINTER,
+			      g_cclosure_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
 			      1,
 			      G_TYPE_POINTER);
@@ -152,7 +147,7 @@ vinagre_fav_class_init (VinagreFavClass *klass)
 			      G_SIGNAL_RUN_FIRST,
 			      G_STRUCT_OFFSET (VinagreFavClass, fav_selected),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__POINTER,
+			      g_cclosure_marshal_VOID__OBJECT,
 			      G_TYPE_NONE,
 			      1,
 			      G_TYPE_POINTER);
@@ -167,13 +162,13 @@ vinagre_fav_row_activated_cb (GtkTreeView       *treeview,
 			      VinagreFav         *fav)
 {
   GtkTreeIter iter;
-  VinagreConnection *conn = NULL;
+  VinagreBookmarksEntry *entry;
   gboolean folder, group;
 
   gtk_tree_model_get_iter (fav->priv->model, &iter, path);
   gtk_tree_model_get (fav->priv->model, 
 		      &iter,
-		      CONN_COL, &conn,
+		      ENTRY_COL, &entry,
                       IS_FOLDER_COL, &folder,
                       IS_GROUP_COL, &group,
 		      -1);
@@ -191,7 +186,7 @@ vinagre_fav_row_activated_cb (GtkTreeView       *treeview,
   g_signal_emit (G_OBJECT (fav), 
 		 signals[FAV_ACTIVATED],
 		 0, 
-		 conn);
+		 entry);
 }
 
 static void
@@ -199,26 +194,26 @@ vinagre_fav_selection_changed_cb (GtkTreeSelection *selection,
 				  VinagreFav       *fav)
 {
   GtkTreeIter iter;
-  VinagreConnection *conn = NULL;
+  VinagreBookmarksEntry *entry;
   gboolean avahi;
 
   if (gtk_tree_selection_get_selected (selection, NULL, &iter))
     {
       gtk_tree_model_get (fav->priv->model, 
 			  &iter,
-			  CONN_COL, &conn,
+			  ENTRY_COL, &entry,
                           IS_AVAHI_COL, &avahi,
 			  -1);
     }
 
   if (avahi)
-    conn = NULL;
+    entry = NULL;
 
   /* Emits the signal saying that user has selected a bookmark */
   g_signal_emit (G_OBJECT (fav), 
 		 signals[FAV_SELECTED],
 		 0, 
-		 conn);
+		 entry);
 }
 
 static GtkTreePath *
@@ -281,13 +276,14 @@ menu_position (GtkMenu    *menu,
 }
 
 static gboolean
-show_popup_menu (VinagreFav     *fav,
-		 GdkEventButton *event)
+show_popup_menu (VinagreFav      *fav,
+		 GdkEventButton  *event,
+		 const gchar     *menu_name)
 {
   GtkWidget *menu = NULL;
 
   menu = gtk_ui_manager_get_widget (vinagre_window_get_ui_manager (fav->priv->window),
-				    "/FavPopup");
+				    menu_name);
   g_return_val_if_fail (menu != NULL, FALSE);
 
   if (event)
@@ -309,9 +305,6 @@ show_popup_menu (VinagreFav     *fav,
 		      fav,
 		      0, 
 		      gtk_get_current_event_time ());
-
-      gtk_menu_shell_select_first (GTK_MENU_SHELL (menu), FALSE);
-
     }
 
   return TRUE;
@@ -323,45 +316,46 @@ fav_button_press_event (GtkTreeView    *treeview,
 			VinagreFav     *fav)
 {
   GtkTreePath *path = NULL;
-  GtkTreeIter iter;
-  gboolean folder, group, avahi;
+  GtkTreeIter  iter;
+  gboolean     folder, group, avahi;
 
   if (!((GDK_BUTTON_PRESS == event->type) && (3 == event->button)))
     return FALSE;
 
-  if (event->window == gtk_tree_view_get_bin_window (treeview))
-    {
-    /* Change the cursor position */
-    if (gtk_tree_view_get_path_at_pos (treeview,
-				       event->x,
-				       event->y,
-				       &path,
-				       NULL,
-				       NULL,
-			               NULL))
-      {				
-        gtk_tree_model_get_iter (fav->priv->model, &iter, path);
-        gtk_tree_model_get (fav->priv->model, 
-		            &iter,
-                            IS_FOLDER_COL, &folder,
-                            IS_GROUP_COL, &group,
-                            IS_AVAHI_COL, &avahi,
-		            -1);
-        if (folder || group || avahi)
-          {
-            gtk_tree_path_free (path);
-            return FALSE;
-          }
+  if (event->window != gtk_tree_view_get_bin_window (treeview))
+    return FALSE;
 
-        gtk_tree_view_set_cursor (treeview,
-				  path,
-				  NULL,
-				  FALSE);
-	 gtk_tree_path_free (path);
-	 /* A row exists at mouse position */
-	 return show_popup_menu (fav, event);
-      }
+  if (gtk_tree_view_get_path_at_pos (treeview,
+				     event->x,
+				     event->y,
+				     &path,
+				     NULL,
+				     NULL,
+			             NULL))
+    {				
+      gtk_tree_model_get_iter (fav->priv->model, &iter, path);
+      gtk_tree_model_get (fav->priv->model, 
+			  &iter,
+			  IS_FOLDER_COL, &folder,
+			  IS_GROUP_COL, &group,
+			  IS_AVAHI_COL, &avahi,
+			  -1);
+      if (group || avahi)
+	{
+	  gtk_tree_path_free (path);
+	  return FALSE;
+	}
+
+      gtk_tree_view_set_cursor (treeview,
+				path,
+				NULL,
+				FALSE);
+      gtk_tree_path_free (path);
+      return show_popup_menu (fav, event, folder?"/FavPopupFolder":"/FavPopupConn");
     }
+  else
+    return show_popup_menu (fav, event, "/FavPopupEmpty");
+
   return FALSE;
 }
 
@@ -372,23 +366,26 @@ fav_popup_menu (GtkWidget  *treeview,
 {
   /* Only respond if the treeview is the actual focus */
   if (gtk_window_get_focus (GTK_WINDOW (fav->priv->window)) == treeview)
-    return show_popup_menu (fav, NULL);
+    return show_popup_menu (fav, NULL, FALSE);
 
   return FALSE;
 }
 
 static gboolean
 vinagre_fav_tooltip (GtkWidget *widget,
-                     gint x, gint y, gboolean k,
+                     gint       x,
+                     gint       y,
+                     gboolean   k,
                      GtkTooltip *tooltip,
                      VinagreFav *fav)
 {
-  gchar *tip, *name, *uri;
-  GtkTreePath *path = NULL;
-  gint bx, by;
-  GtkTreeIter iter;
-  gboolean folder, group;
-  VinagreConnection *conn = NULL;
+  gchar                 *tip, *name, *uri;
+  GtkTreePath           *path;
+  gint                   bx, by;
+  GtkTreeIter            iter;
+  gboolean               folder, group;
+  VinagreConnection     *conn;
+  VinagreBookmarksEntry *entry;
 
   gtk_tree_view_convert_widget_to_bin_window_coords (GTK_TREE_VIEW (widget),
                                                      x, y,
@@ -406,7 +403,7 @@ vinagre_fav_tooltip (GtkWidget *widget,
   gtk_tree_model_get_iter (fav->priv->model, &iter, path);
   gtk_tree_model_get (fav->priv->model, 
 		      &iter,
-		      CONN_COL, &conn, 
+		      ENTRY_COL, &entry, 
                       IS_FOLDER_COL, &folder,
                       IS_GROUP_COL, &group,
 		      -1);
@@ -416,6 +413,7 @@ vinagre_fav_tooltip (GtkWidget *widget,
   if (folder || group)
     return FALSE;
 
+  conn = vinagre_bookmarks_entry_get_conn (entry);
   name = vinagre_connection_get_best_name (conn);
   uri = vinagre_connection_get_string_rep (conn, TRUE);
   tip = g_markup_printf_escaped ("<b>%s</b>\n"
@@ -431,25 +429,6 @@ vinagre_fav_tooltip (GtkWidget *widget,
   return TRUE;
 }
 
-static void
-vinagre_fav_indent_level1_cell_data_func (GtkTreeViewColumn *tree_column,
-					  GtkCellRenderer   *cell,
-					  GtkTreeModel      *model,
-					  GtkTreeIter       *iter,
-					  VinagreFav        *fav)
-{
-  GtkTreePath *path;
-  int          depth;
-
-  path = gtk_tree_model_get_path (model, iter);
-  depth = gtk_tree_path_get_depth (path);
-  gtk_tree_path_free (path);
-  g_object_set (cell,
-	        "text", " ",
-		"visible", depth > 1,
-		NULL);
-}
-
 static void
 vinagre_fav_cell_set_background (VinagreFav       *fav,
 				 GtkCellRenderer  *cell,
@@ -559,47 +538,13 @@ vinagre_fav_title_cell_data_func (GtkTreeViewColumn *column,
   g_free (str);
 }
 
-static void
-vinagre_fav_expander_cell_data_func (GtkTreeViewColumn *column,
-				     GtkCellRenderer   *cell,
-				     GtkTreeModel      *model,
-				     GtkTreeIter       *iter,
-				     VinagreFav        *fav)
-{
-  gboolean is_group;
-  gboolean is_active;
-
-  gtk_tree_model_get (model, iter,
-		      IS_GROUP_COL, &is_group,
-		      -1);
-
-  if (gtk_tree_model_iter_has_child (model, iter))
-    {
-      GtkTreePath *path;
-      gboolean     row_expanded;
-
-      path = gtk_tree_model_get_path (model, iter);
-      row_expanded = gtk_tree_view_row_expanded (GTK_TREE_VIEW (column->tree_view), path);
-      gtk_tree_path_free (path);
-
-      g_object_set (cell,
-		    "visible", TRUE,
-		    "expander-style", row_expanded ? GTK_EXPANDER_EXPANDED : GTK_EXPANDER_COLLAPSED,
-		    NULL);
-    }
-  else
-    g_object_set (cell, "visible", FALSE, NULL);
-
-  is_active = FALSE;
-  vinagre_fav_cell_set_background (fav, cell, is_group, is_active);
-}
-
 static void
 vinagre_fav_create_tree (VinagreFav *fav)
 {
   GtkCellRenderer   *cell;
   GtkWidget         *scroll;
   GtkTreeSelection  *selection;
+  GtkTreeViewColumn *main_column;
 
   /* Create the scrolled window */
   scroll = gtk_scrolled_window_new (NULL, NULL);
@@ -616,71 +561,42 @@ vinagre_fav_create_tree (VinagreFav *fav)
   fav->priv->model = GTK_TREE_MODEL (gtk_tree_store_new (NUM_COLS,
 							 GDK_TYPE_PIXBUF,
 							 G_TYPE_STRING,
-							 VINAGRE_TYPE_CONNECTION,
+							 VINAGRE_TYPE_BOOKMARKS_ENTRY,
                                                          G_TYPE_BOOLEAN,
                                                          G_TYPE_BOOLEAN,
                                                          G_TYPE_BOOLEAN));
 
   fav->priv->tree = gtk_tree_view_new_with_model (fav->priv->model);
-  g_object_set (fav->priv->tree, "show-expanders", FALSE, NULL);
   gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (fav->priv->tree), FALSE);
   selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (fav->priv->tree));
   gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
   g_signal_connect (selection, "changed", G_CALLBACK (vinagre_fav_selection_changed_cb), fav);
 
-  fav->priv->main_column = gtk_tree_view_column_new ();
-//  gtk_tree_view_column_set_title (fav->priv->main_column, _("S_ource"));
-  gtk_tree_view_column_set_clickable (fav->priv->main_column, FALSE);
-  gtk_tree_view_append_column (GTK_TREE_VIEW (fav->priv->tree), fav->priv->main_column);
-
-  /* Set up the indent level1 column */
-  cell = gtk_cell_renderer_text_new ();
-  gtk_tree_view_column_pack_start (fav->priv->main_column, cell, FALSE);
-  gtk_tree_view_column_set_cell_data_func (fav->priv->main_column,
-					   cell,
-				           (GtkTreeCellDataFunc) vinagre_fav_indent_level1_cell_data_func,
-				           fav,
-					   NULL);
-  g_object_set (cell,
-		"xpad", 0,
-		"visible", FALSE,
-		NULL);
+  main_column = gtk_tree_view_column_new ();
+  gtk_tree_view_column_set_clickable (main_column, FALSE);
+  gtk_tree_view_append_column (GTK_TREE_VIEW (fav->priv->tree), main_column);
 
   /* Set up the pixbuf column */
   cell = gtk_cell_renderer_pixbuf_new ();
-  gtk_tree_view_column_pack_start (fav->priv->main_column, cell, FALSE);
-  gtk_tree_view_column_set_cell_data_func (fav->priv->main_column,
+  gtk_tree_view_column_pack_start (main_column, cell, FALSE);
+  gtk_tree_view_column_set_cell_data_func (main_column,
 					   cell,
 					   (GtkTreeCellDataFunc) vinagre_fav_pixbuf_cell_data_func,
 					    fav,
 					    NULL);
-  g_object_set (cell,
-		"xpad", 8,
-		"ypad", 1,
-		"visible", FALSE,
-		NULL);
 
   /* Set up the name column */
   cell = gtk_cell_renderer_text_new ();
   g_object_set (cell,
 		"ellipsize", PANGO_ELLIPSIZE_END,
 		 NULL);
-  gtk_tree_view_column_pack_start (fav->priv->main_column, cell, TRUE);
-  gtk_tree_view_column_set_cell_data_func (fav->priv->main_column,
+  gtk_tree_view_column_pack_start (main_column, cell, TRUE);
+  gtk_tree_view_column_set_cell_data_func (main_column,
 				           cell,
 					   (GtkTreeCellDataFunc) vinagre_fav_title_cell_data_func,
 					    fav,
 					    NULL);
 
-  /* Expander */
-  cell = gossip_cell_renderer_expander_new ();
-  gtk_tree_view_column_pack_end (fav->priv->main_column, cell, FALSE);
-  gtk_tree_view_column_set_cell_data_func (fav->priv->main_column,
-					   cell,
-				           (GtkTreeCellDataFunc) vinagre_fav_expander_cell_data_func,
-                                           fav,
-					   NULL);
-
   g_signal_connect (fav->priv->tree,
 		    "row-activated",
 		    G_CALLBACK (vinagre_fav_row_activated_cb),
@@ -710,6 +626,96 @@ vinagre_fav_create_tree (VinagreFav *fav)
   g_object_unref (G_OBJECT (fav->priv->model));
 }
 
+static void
+drag_data_received_handl (GtkWidget *widget,
+			  GdkDragContext *context,
+			  gint x,
+			  gint y,
+			  GtkSelectionData *selection_data,
+			  guint target_type,
+			  guint time,
+			  VinagreFav *fav)
+{
+  gchar *_sdata;
+  gboolean success = FALSE;
+
+  if ((selection_data != NULL) && (selection_data->length >= 0))
+    {
+      switch (target_type)
+	{
+	  case TARGET_VINAGRE:
+	    _sdata = (gchar*)selection_data-> data;
+	    success = TRUE;
+	   break;
+
+	  default:
+	    g_print ("nothing good");
+	}
+    }
+
+  gtk_drag_finish (context, success, FALSE, time);
+
+  if (success)
+    {
+      VinagreConnection *conn;
+      gchar **info;
+      gchar *error = NULL;
+
+      info = g_strsplit (_sdata, "||", 2);
+      if (g_strv_length (info) != 2)
+	{
+	  vinagre_utils_show_error (_("Invalid operation"),
+				    _("Data received from drag&drop operation is invalid."),
+				    GTK_WINDOW (fav->priv->window));
+	  return;
+	}
+
+      conn = vinagre_connection_new_from_string (info[1], &error, FALSE);
+      if (!conn)
+	{
+	  g_strfreev (info);
+	  vinagre_utils_show_error (NULL,
+				    error ? error : _("Unknown error"),
+				    GTK_WINDOW (fav->priv->window));
+	  return;
+	}
+
+      vinagre_connection_set_name (conn, info[0]);
+      vinagre_bookmarks_add (vinagre_bookmarks_get_default (),
+			     conn,
+			     GTK_WINDOW (fav->priv->window));
+
+      g_strfreev (info);
+      g_object_unref (conn);
+    }
+}
+
+static gboolean
+drag_drop_handl (GtkWidget *widget,
+		 GdkDragContext *context,
+		 gint x,
+		 gint y,
+		 guint time,
+		 gpointer user_data)
+{
+  gboolean is_valid_drop_site;
+  GdkAtom  target_type;
+
+  is_valid_drop_site = FALSE;
+
+  if (context->targets)
+    {
+      target_type = GDK_POINTER_TO_ATOM (g_list_nth_data (context->targets, TARGET_VINAGRE));
+      gtk_drag_get_data (widget,
+			 context,
+			 target_type,
+			 time);
+      is_valid_drop_site = TRUE;
+    }
+
+  return  is_valid_drop_site;
+}
+
 static void
 vinagre_fav_init (VinagreFav *fav)
 {
@@ -720,14 +726,31 @@ vinagre_fav_init (VinagreFav *fav)
   /* setup the tree */
   vinagre_fav_create_tree (fav);
 
+  gtk_drag_dest_set (fav->priv->tree,
+		     GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_HIGHLIGHT,
+		     vinagre_target_list,
+		     vinagre_n_targets,
+		     GDK_ACTION_COPY);
+
+  g_signal_connect (fav->priv->tree,
+		    "drag-data-received",
+		    G_CALLBACK(drag_data_received_handl),
+		    fav);
+  g_signal_connect (fav->priv->tree,
+		    "drag-drop",
+		    G_CALLBACK (drag_drop_handl),
+		    NULL);
+
   g_signal_connect_swapped (vinagre_bookmarks_get_default (),
                             "changed",
                             G_CALLBACK (vinagre_fav_update_list),
                             fav);
+#ifdef VINAGRE_ENABLE_AVAHI
   g_signal_connect_swapped (vinagre_mdns_get_default (),
                             "changed",
                             G_CALLBACK (vinagre_fav_update_list),
                             fav);
+#endif
 }
 
 GtkWidget *
@@ -740,6 +763,68 @@ vinagre_fav_new (VinagreWindow *window)
 				   NULL));
 }
 
+static void
+vinagre_fav_fill_bookmarks (GtkTreeStore *store, GSList *list, GtkTreeIter *parent_iter, gboolean is_avahi)
+{
+  GdkPixbuf             *pixbuf;
+  GtkIconTheme          *icon_theme;
+  gchar                 *name;
+  VinagreBookmarksEntry *entry;
+  GSList                *l;
+  GtkTreeIter            iter;
+
+ for (l = list; l; l = l->next)
+    {
+      entry = (VinagreBookmarksEntry *) l->data;
+      switch (vinagre_bookmarks_entry_get_node (entry))
+	{
+	  case VINAGRE_BOOKMARKS_ENTRY_NODE_FOLDER:
+	    icon_theme = gtk_icon_theme_get_default ();
+	    pixbuf = gtk_icon_theme_load_icon  (icon_theme,
+					        "folder",
+						16,
+						0,
+						NULL);
+
+	    gtk_tree_store_append (store, &iter, parent_iter);
+	    gtk_tree_store_set (store, &iter,
+				IMAGE_COL, pixbuf,
+				NAME_COL, vinagre_bookmarks_entry_get_name (entry),
+				ENTRY_COL, entry,
+				IS_FOLDER_COL, TRUE,
+				IS_GROUP_COL, FALSE,
+				IS_AVAHI_COL, FALSE,
+				-1);
+	    if (pixbuf != NULL)
+	      g_object_unref (pixbuf);
+
+	    vinagre_fav_fill_bookmarks (store, vinagre_bookmarks_entry_get_children (entry), &iter, is_avahi);
+	    break;
+
+	  case VINAGRE_BOOKMARKS_ENTRY_NODE_CONN:
+	    name = vinagre_connection_get_best_name (vinagre_bookmarks_entry_get_conn (entry));
+	    pixbuf = vinagre_connection_get_icon (vinagre_bookmarks_entry_get_conn (entry));
+
+	    gtk_tree_store_append (store, &iter, parent_iter);
+	    gtk_tree_store_set (store, &iter,
+				IMAGE_COL, pixbuf,
+				NAME_COL, name,
+				ENTRY_COL, entry,
+				IS_FOLDER_COL, FALSE,
+				IS_GROUP_COL, FALSE,
+				IS_AVAHI_COL, is_avahi,
+				-1);
+	    g_free (name);
+	    if (pixbuf != NULL)
+	      g_object_unref (pixbuf);
+	    break;
+
+	  default:
+	    g_assert_not_reached ();
+	}
+    }
+}
+
 gboolean
 vinagre_fav_update_list (VinagreFav *fav)
 {
@@ -747,7 +832,6 @@ vinagre_fav_update_list (VinagreFav *fav)
   GtkTreeStore      *store;
   GSList            *list, *l, *next;
   GdkPixbuf         *pixbuf;
-  GtkTreePath       *path;
     
   g_return_val_if_fail (VINAGRE_IS_FAV (fav), FALSE);
 
@@ -765,39 +849,9 @@ vinagre_fav_update_list (VinagreFav *fav)
                       IS_AVAHI_COL, FALSE,
                       -1);
 
-  for (l = list; l; l = next)
-    {
-      gchar *name = NULL;
-      VinagreConnection *conn;
+  vinagre_fav_fill_bookmarks (store, list, NULL, FALSE);
 
-      next = l->next;
-
-      conn = (VinagreConnection *) l->data;
-      g_assert (VINAGRE_IS_CONNECTION (conn));
-
-      name = vinagre_connection_get_best_name (conn);
-      pixbuf = vinagre_connection_get_icon (conn);
-
-      gtk_tree_store_append (store, &iter, &parent_iter);
-      gtk_tree_store_set (store, &iter,
-                          IMAGE_COL, pixbuf,
-                          NAME_COL, name,
-                          CONN_COL, conn,
-                          IS_FOLDER_COL, FALSE,
-                          IS_GROUP_COL, FALSE,
-                          IS_AVAHI_COL, FALSE,
-                          -1);
-      if (name)
-        g_free (name);
-      if (pixbuf != NULL)
-	g_object_unref (pixbuf);
-    }
-
-  path = gtk_tree_path_new_from_string ("0");
-  gtk_tree_view_expand_row (GTK_TREE_VIEW (fav->priv->tree), path, FALSE);
-  gtk_tree_path_free (path);
-
-  /* avahi */
+#ifdef VINAGRE_ENABLE_AVAHI
   list = vinagre_mdns_get_all (vinagre_mdns_get_default ());
   if (!list)
     return FALSE;
@@ -810,38 +864,9 @@ vinagre_fav_update_list (VinagreFav *fav)
                       IS_AVAHI_COL, FALSE,
                       -1);
 
-  for (l = list; l; l = next)
-    {
-      gchar *name = NULL;
-      VinagreConnection *conn;
-
-      next = l->next;
-
-      conn = (VinagreConnection *) l->data;
-      g_assert (VINAGRE_IS_CONNECTION (conn));
-
-      name = vinagre_connection_get_best_name (conn);
-      pixbuf = vinagre_connection_get_icon (conn);
-
-      gtk_tree_store_append (store, &iter, &parent_iter);
-      gtk_tree_store_set (store, &iter,
-                          IMAGE_COL, pixbuf,
-                          NAME_COL, name,
-                          CONN_COL, conn,
-                          IS_FOLDER_COL, FALSE,
-                          IS_GROUP_COL, FALSE,
-                          IS_AVAHI_COL, TRUE,
-                          -1);
-      if (name)
-        g_free (name);
-      if (pixbuf != NULL)
-	g_object_unref (pixbuf);
-    }
-
-  path = gtk_tree_path_new_from_string ("1");
-  gtk_tree_view_expand_row (GTK_TREE_VIEW (fav->priv->tree), path, FALSE);
-  gtk_tree_path_free (path);
+  vinagre_fav_fill_bookmarks (store, list, NULL, TRUE);
+#endif
 
   return FALSE;
 }
-/* vim: ts=8 */
+/* vim: set ts=8: */
